/*----------------------------------------------------------------------*\
 | C routines that read shapefiles using Frank Warmerdam's shapelib		|
 | and make the results available to Tcl procedures.					|
 |	http://pobox.com/~warmerdam											|
 |																		|
 | What do I want to do with shapefiles in Tcl?							|
 |																		|
 | shp d -open $input_file												|
 |		opens shapefile, returns a handle								|
 |																		|
 | $d info [shape_number]												|
 |		if no shape number given, returns {shape_type shape_count}		|
 |		else returns a list containing 									|
 |			shape_type vertex_count part_count							|
 |			and for each part, a list containing part_type part_start	|
 |																		|
 | $d bounds [shape_number]												|
 |		if no shape number given,										|
 |			returns the bounding box of the entire shapefile, as		|
 |			{Xmin Ymin Zmin Mmin Xmax Ymax Zmax Mmax}					|
 |		else returns same info about the requested shape				|
 |																		|
 | $d coords shape_number												|
 |		returns a list of lists (one for each shape part, each list		|
 |		containing coordinates)											|
 |																		|
 | $d forget															|
 |		closes shapefile												|
 |																		|
 | Peter N. Schweitzer (U.S. Geological Survey, Reston VA 20192)		|
\*----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <tcl.h>

#include "shp.h"

#include <shapefil.h>

static char *empty = "";
static char *failure = "0";
static char *success = "1";
static int record_count = 0;
static char message[512];

int process_shp_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	int i,j,k,n;
	SHPHandle sf = (SHPHandle) clientData;
	Tcl_Obj *obj;
	int fc,rc;
	char *s;
	char number[1024];

	int nShapeType, nEntities, iPart;
	const char *type_name;
	double adfMinBound[4], adfMaxBound[4];
	SHPObject *psShape;

	if (argc > 1) {
		const char *command = argv[1];

		/*--------------------------------------------------------------*\
		 | info [shape_number]
		\*--------------------------------------------------------------*/

		if (strcmp (command,"info") == 0) {

			if (!sf) {
				Tcl_SetResult (interp,"info: cannot find; no shapefile has been read",TCL_STATIC);
				return (TCL_ERROR);
				}

			if (argc > 2) {
				char *rowid = argv[2];
				char *t;
				i = (int) strtoul (rowid,&t,0);
				if (t == rowid) {
					sprintf (message,"bounds: cannot interpret shape number from \"%s\"",rowid);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				if (psShape = SHPReadObject (sf,i)) {
					Tcl_ResetResult (interp);
					obj = Tcl_NewListObj (0,NULL);

					type_name = SHPTypeName (psShape->nSHPType);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (type_name,strlen(type_name)));
					sprintf (number,"%lf",psShape->nVertices);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->nParts);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));

					for (j=0; j < psShape->nParts; j++) {
						Tcl_Obj *obj2;
						obj2 = Tcl_NewListObj (0,NULL);

						type_name = SHPPartTypeName (psShape->panPartType[j]);
						Tcl_ListObjAppendElement (interp,obj2,Tcl_NewStringObj (type_name,strlen(type_name)));

						sprintf (number,"%d",psShape->panPartStart[j]);
						Tcl_ListObjAppendElement (interp,obj2,Tcl_NewStringObj (number,strlen(number)));

						Tcl_ListObjAppendElement (interp,obj,obj2);
						}
					Tcl_SetObjResult (interp,obj);
					return (TCL_OK);
					}
				else {
					sprintf (message,"Error: no shape number %d exists",i);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				SHPGetInfo (sf, &nEntities, &nShapeType, adfMinBound, adfMaxBound);

				Tcl_ResetResult (interp);
				obj = Tcl_NewListObj (0,NULL);

				type_name = SHPTypeName (nShapeType);
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (type_name,strlen(type_name)));

				sprintf (number,"%d",nEntities);
				Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));

				Tcl_SetObjResult (interp,obj);
				return (TCL_OK);
				}
			}

		/*--------------------------------------------------------------*\
		 | bounds [shape_number]
		\*--------------------------------------------------------------*/

		if (strcmp (command,"bounds") == 0) {

			if (!sf) {
				Tcl_SetResult (interp,"fields: cannot find; no shapefile has been read",TCL_STATIC);
				return (TCL_ERROR);
				}

			if (argc > 2) {
				char *rowid = argv[2];
				char *t;
				i = (int) strtoul (rowid,&t,0);
				if (t == rowid) {
					sprintf (message,"bounds: cannot interpret shape number from \"%s\"",rowid);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				if (psShape = SHPReadObject (sf,i)) {
					Tcl_ResetResult (interp);
					obj = Tcl_NewListObj (0,NULL);

					sprintf (number,"%lf",psShape->dfXMin);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfYMin);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfZMin);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfMMin);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfXMax);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfYMax);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfZMax);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					sprintf (number,"%lf",psShape->dfMMax);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));

					Tcl_SetObjResult (interp,obj);
					return (TCL_OK);
					}
				else {
					sprintf (message,"Error: no shape number %d exists",i);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				SHPGetInfo (sf, &nEntities, &nShapeType, adfMinBound, adfMaxBound);

				Tcl_ResetResult (interp);
				obj = Tcl_NewListObj (0,NULL);
				for (j=0; j < 4; j++) {
					sprintf (number,"%lf",adfMinBound[j]);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					}
				for (j=0; j < 4; j++) {
					sprintf (number,"%lf",adfMaxBound[j]);
					Tcl_ListObjAppendElement (interp,obj,Tcl_NewStringObj (number,strlen(number)));
					}
				Tcl_SetObjResult (interp,obj);
				return (TCL_OK);
				}
			}

		/*--------------------------------------------------------------*\
		 | coords shape_number
		\*--------------------------------------------------------------*/

		if (strcmp (command,"coords") == 0)
			if (argc > 2) {
				char *rowid = argv[2];
				char *t;

				if (!sf) {
					Tcl_SetResult (interp,"coords cannot find; no shapefile has been read",TCL_STATIC);
					return (TCL_ERROR);
					}

				i = (int) strtoul (rowid,&t,0);
				if (t == rowid) {
					sprintf (message,"coords: cannot interpret shape number from \"%s\"",rowid);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}

				if (psShape = SHPReadObject (sf,i)) {
					Tcl_ResetResult (interp);
					obj = Tcl_NewListObj (0,NULL);

					if (psShape->nVertices > 0) {

						/*----------------------------------------------*\
						 | If the shape says it has zero parts, then it	|
						 | really has one part, and the vertices are	|
						 | given simply.  If there are more than zero	|
						 | parts, then the start indexes of the parts	|
						 | are specified in a list (panPartStart[]), so	|
						 | this array of indexes is used to keep the	|
						 | vertices of each part separate.				|
						\*----------------------------------------------*/

						if (psShape->nParts > 0) {
							for (j=0; j < psShape->nParts; j++) {
								Tcl_Obj *obj2;
								obj2 = Tcl_NewListObj (0,NULL);

								if (j < psShape->nParts - 1) k = psShape->panPartStart[j+1];
								else k = psShape->nVertices;

								for (n=psShape->panPartStart[j]; n < k; n++) {
									Tcl_Obj *obj3;
									obj3 = Tcl_NewListObj (0,NULL);

									sprintf (number,"%lf",psShape->padfX[n]);
									Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));
									sprintf (number,"%lf",psShape->padfY[n]);
									Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));

									if (psShape->nSHPType > 10) {
										sprintf (number,"%lf",psShape->padfZ[n]);
										Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));
										sprintf (number,"%lf",psShape->padfM[n]);
										Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));
										}

									Tcl_ListObjAppendElement (interp,obj2,obj3);
									}

								Tcl_ListObjAppendElement (interp,obj,obj2);
								}
							}
						else {
							Tcl_Obj *obj2;
							obj2 = Tcl_NewListObj (0,NULL);

							/*------------------------------------------*\
							 | The simple case; vertices are all of one	|
							 | part, and there is no index array.		|
							\*------------------------------------------*/

							for (n=0; n < psShape->nVertices; n++) {
								Tcl_Obj *obj3;
								obj3 = Tcl_NewListObj (0,NULL);

								sprintf (number,"%lf",psShape->padfX[n]);
								Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));
								sprintf (number,"%lf",psShape->padfY[n]);
								Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));

								if (psShape->nSHPType > 10) {
									sprintf (number,"%lf",psShape->padfZ[n]);
									Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));
									sprintf (number,"%lf",psShape->padfM[n]);
									Tcl_ListObjAppendElement (interp,obj3,Tcl_NewStringObj (number,strlen(number)));
									}

								Tcl_ListObjAppendElement (interp,obj2,obj3);
								}
							Tcl_ListObjAppendElement (interp,obj,obj2);
							}
						}

					Tcl_SetObjResult (interp,obj);
					return (TCL_OK);
					}
				else {
					sprintf (message,"Error: no shape number %d exists",i);
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			else {
				Tcl_SetResult (interp,"coords expects the number of a record",TCL_STATIC);
				return (TCL_ERROR);
				}

		/*--------------------------------------------------------------*\
		 | forget
		\*--------------------------------------------------------------*/

		if (strcmp (command,"forget") == 0) {
			if (sf) {
				SHPClose (sf);
				Tcl_DeleteCommand (interp,argv[0]);
				Tcl_SetResult (interp,success,TCL_STATIC);
				}
			else
				Tcl_SetResult (interp,failure,TCL_STATIC);
			return (TCL_OK);
			}

		}
	return (TCL_OK);
	}

int shp_cmd (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
	char *variable_name;
	char *input_file;
	char id [64];
	char *s;
	char *text_buffer = NULL;
	int i;
	SHPHandle sf;

	Tcl_ResetResult (interp);

	if (argc > 1) {
		variable_name = argv[1];
		if (argc > 1) {

			/*----------------------------------------------------------*\
			 | -open input_file
			\*----------------------------------------------------------*/

			if (strcmp (argv[2],"-open") == 0) {
				if (argc > 2) {
					input_file = argv[3];

					/*--------------------------------------------------*\
					 | Open the input file creating a new command.		|
					\*--------------------------------------------------*/

					if (input_file)
						if (sf = SHPOpen (input_file,"rb")) {
							sprintf (id,"shp.%04X",record_count++);
							Tcl_SetVar (interp,variable_name,id,0);
							Tcl_CreateCommand (interp,id,(Tcl_CmdProc *) process_shp_cmd,(ClientData)sf, (Tcl_CmdDeleteProc *)NULL);
							Tcl_SetResult (interp,success,TCL_STATIC);
							}
						else
							Tcl_SetResult (interp,failure,TCL_STATIC);
					else
						Tcl_SetResult (interp,failure,TCL_STATIC);

					return (TCL_OK);
					}
				else {
					sprintf (message,"Error: no input file name given");
					Tcl_SetResult (interp,message,TCL_STATIC);
					return (TCL_ERROR);
					}
				}
			}
		else {
			sprintf (message,"Error: got variable name %s, no -open",argv[1]);
			Tcl_SetResult (interp,message,TCL_STATIC);
			return (TCL_ERROR);
			}
		}
	else {
		Tcl_SetResult (interp,"shp expects variable name",TCL_STATIC);
		return (TCL_ERROR);
		}
	return (TCL_OK);
	}


int Shp_Init (Tcl_Interp *interp) {
	if (Tcl_InitStubs (interp,"8.1",0) == NULL) return (TCL_ERROR);
	if (Tcl_PkgRequire (interp,"Tcl","8.0",0) == NULL) return (TCL_ERROR);
	if (Tcl_PkgProvide (interp,"shp","1.0.2") != TCL_OK) return (TCL_ERROR);

	Tcl_CreateCommand (interp,"shp",(Tcl_CmdProc *) shp_cmd, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
	return (TCL_OK);
	}

/*----------------------------------------------------------------------*\
\*----------------------------------------------------------------------*/
